function [MA_]=nonlinear_MA(M_,oo_,options_,varargin)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% nonlinear_MA.m
%
% This is the main program of the package nonlinear_MA
% 
% Note that the folder containing the package must be included in your path
%
% To run, first run dynare with a .mod file and then enter at the command 
% prompt or place at the end of .mod file the following command
%
%  nonlinear_MA(M_,oo_,options_)
%
% The inputs
% M_
% oo_
% options_
% are produced by dynare.
%
% If you would like to change some of the defaults, run 
% create a structured  array with fields named commesurately with the names
% found below in the section % setup parameters. Pass the structured array
% as a fourth input when calling the program:
%
% nonlinear_MA(M_,oo_,options_, name)
%
% where "name" is whatever name you give to your structured array. For
% example, to change the default of a postive one standard degree shock for
% impulse response to minus three standard degree shock, type 
% >> myoptions.shock_scale=-3;
% and then call nonlinear_MA
% >> nonlinear_MA(M_,oo_,options_, myoptions)
%
%THIS VERSION: 1.0.1 October 12, 201
%
%Copyright: Hong Lan and Alexander Meyer-Gohde
%
%You are free to use/modify/redistribute this program so long as original
%authorship credit is given and you in no way impinge on its free
%distribution
%This software is provided as is with no guarantees of any kind
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

tic;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% setup parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if nargin > 4 %Improper number of inputs
        disp('Too many inputs passed to nonlinear_MA')
else %Otherwise correct number of inputs.
    %I'll start by setting the defaults. If a structure array was passed, 
    %I'll read for override values afterwards
    MA_.shock_scale=1; %Positive one standard deviation shock in IRFs
    if options_.periods>0
        MA_.calculate_simulations=1;
        MA_.plot_simulations=1;
        MA_.simulation_nbr=1; %Once the kernels have been calculated, 
                                %multiple sims could be run
    else
        MA_.calculate_simulations=0;
    end
    if options_.irf>0
        MA_.calculate_irf=1;
        MA_.plot_irf=1;
    else
        MA_.calculate_irf=0;
        MA_.plot_irf=0;
    end
    MA_.calculate_kernels=0;
    MA_.plot_kernels=0;
    MA_.max_kernel_length=40; 
    if nargin==4 %This is where I'll start overriding the defaults, given
            %that the user passed a strutural array to the program
            %containing override values
        MA_options = varargin{1};
        if isfield(MA_options, 'shock_scale')==1
            MA_.shock_scale=MA_options.shock_scale;
        end
        if isfield(MA_options, 'max_kernel_length')==1
            MA_.max_kernel_length=MA_options.max_kernel_length;
        end
        if isfield(MA_options, 'calculate_kernels')==1
            MA_.calculate_kernels=MA_options.calculate_kernels;
        end
        if isfield(MA_options, 'plot_kernels')==1
            MA_.plot_kernels=MA_options.plot_kernels;
        end
        if isfield(MA_options, 'calculate_simulations')==1
            MA_.calculate_simulations=MA_options.calculate_simulations;
        end
        if isfield(MA_options, 'plot_simulations')==1
            MA_.plot_simulations=MA_options.calculate_simulations;
        end
        if isfield(MA_options, 'simulation_nbr')==1
            MA_.simulation_nbr=MA_options.simulation_nbr;
        end
        if isfield(MA_options, 'calculate_irf')==1
            MA_.calculate_irf=MA_options.calculate_irf;
        end
        if isfield(MA_options, 'plot_irf')==1
            MA_.plot_irf=MA_options.plot_irf;
        end
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% collect model information
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

MA_.klen = M_.maximum_endo_lag + M_.maximum_endo_lead + 1;
MA_.iyv  = M_.lead_lag_incidence';
MA_.iyv  = MA_.iyv(:);
it_= M_.maximum_lag + 1;

% prepare steady state values
ma_steady_state_full_length = [oo_.dr.ys;oo_.dr.ys;oo_.dr.ys];
ma_steady_state             = ma_steady_state_full_length(MA_.iyv>0);

% evaluate jacobian and hessian at steady state
[waste,MA_.df1,MA_.df2,MA_.df3] = feval( [M_.fname '_dynamic'], ma_steady_state, [zeros(it_+1,M_.exo_nbr) ],  M_.params, oo_.dr.ys, it_);





  
%We want to create a rectangular matrix g_picker that maps from all the
%variables to just those with derivatives available in Dynare (variables that actually show up at some lead or lag). Thus, we
%seek a rectangular matrix (# of non zero by # of all variables) with a
%one at the location j,i if the j'th varaible in Dynare corresponds to the
%i'th varaible in the full list.
MA_.non_zero_variables=ones(M_.endo_nbr*MA_.klen+M_.exo_nbr,1);MA_.non_zero_variables(MA_.iyv==0)=0; %Here, I first make a vector with as many ones as total varaibles and then set the elements to zero that correspond to the variables that Dynare zeros out.
MA_.g_picker = spdiags(MA_.non_zero_variables,0,M_.endo_nbr*MA_.klen+M_.exo_nbr,M_.endo_nbr*MA_.klen+M_.exo_nbr); %I turn this into a diagonal matrix, with ones corresponding to variables that Dynare keeps
MA_.g_picker(sum(MA_.g_picker,2)==0,:)=[];%I squash the square diagonal matrix into a rectangular matrix, keeping only rows with a one somewhere (i.e., variables that Dynare provides derivatives for).
MA_.non_zero_variables=sum(MA_.non_zero_variables); %The variables now gives the number of non zero variables. Useful in subsequent calculations

disp('Computing First-Order Terms')
MA_=first_order(M_,MA_);
disp('Done!')
if options_.order>=2
    disp('Computing Second-Order Terms')
    MA_=second_order(M_,MA_);
    disp('Done!')
    if options_.order==3
        disp('Computing Third-Order Terms')
        if M_.endo_nbr < 6 && M_.exo_nbr <3
            try
                MA_=third_order(M_,MA_);
            catch
                MA_=third_order_sparse(M_,MA_);
            end
        else
            MA_=third_order_sparse(M_,MA_);
        end
        disp('Done!')
    end
end

if MA_.calculate_kernels==1
    disp('Computing Kernels. This may take a while')
    [MA_]=kernels(M_,MA_,options_);
    disp('Done!')
    if MA_.plot_kernels==1
    [MA_]=plots_kernels(M_,MA_,options_);
    end
end
if MA_.calculate_simulations==1
    disp('Computing Simulations. This may take a while')
    [MA_]=MA_simul(M_,MA_,options_,oo_);
    disp('Done!')
end
if MA_.plot_irf==1
        disp('Computing Impulse Responses')
        [MA_]=calculate_irfs(M_,MA_,options_,oo_);
         disp('Done!')
        [MA_]=plots_irf(M_,MA_,options_,oo_);
end
time_to_run=toc;
hours_to_run=floor(time_to_run/3600);
minutes_to_run=floor((time_to_run-hours_to_run*3600)/60);
seconds_to_run=time_to_run-hours_to_run*3600-minutes_to_run*60;
disp(['Execution Time:' blanks(2) num2str(hours_to_run) ' hour(s), ' num2str(minutes_to_run) ' minute(s), and '  num2str(seconds_to_run) ' second(s)'])
    